#!.venv/bin/python

import os
import subprocess
import sys
import traceback
import unittest

from test.runtime import ResultAdapter, StyledStream


if __name__ == "__main__":
    successful = False
    stream = sys.stdout
    styled = StyledStream(stream)

    def println(s: str = "") -> None:
        if s:
            stream.write(s)
        stream.write("\n")
        stream.flush()

    try:
        println(styled.h1("§1 Setup"))
        println(styled.h2("Python"))
        println(f"{sys.executable}")
        println(styled.h2("Python Path"))
        for path in sys.path:
            println(f"{path}")
        println(styled.h2("Current Directory"))
        println(f"{os.getcwd()}")
        println(styled.h2("Current Module"))
        println(f"{__file__}")

        println(styled.h1("§2  Type Checking"))
        subprocess.run(["./node_modules/.bin/pyright"], check=True)

        println(styled.h1("§3  Unit Testing"))
        runner = unittest.main(
            module="test",
            exit=False,
            testRunner=unittest.TextTestRunner(
                stream=stream, resultclass=ResultAdapter
            ),
        )
        successful = runner.result.wasSuccessful()

    except subprocess.CalledProcessError:
        println(styled.failure("demicode failed to type check!"))
        exit(1)
    except Exception as x:
        trace = traceback.format_exception(x)
        println("".join(trace[:-1]))
        println(styled.err(trace[-1]))

    sys.exit(not successful)



"""Conversion between color formats and spaces"""
import itertools
import math
from typing import cast, TypeAlias

from .spec import ConverterSpec, CoordinateSpec


_RGB6_TO_RGB256 = (0, 0x5F, 0x87, 0xAF, 0xD7, 0xFF)

# See https://github.com/color-js/color.js/blob/a77e080a070039c534dda3965a769675aac5f75e/src/spaces/srgb-linear.js

_XYZ_TO_LINEAR_SRGB = (
	(  3.2409699419045226,  -1.537383177570094,   -0.4986107602930034  ),
	( -0.9692436362808796,   1.8759675015077202,   0.04155505740717559 ),
	(  0.05563007969699366, -0.20397695888897652,  1.0569715142428786  ),
)

_LINEAR_SRGB_TO_XYZ = (
	( 0.41239079926595934, 0.357584339383878,   0.1804807884018343  ),
	( 0.21263900587151027, 0.715168678767756,   0.07219231536073371 ),
	( 0.01933081871559182, 0.11919477979462598, 0.9505321522496607  ),
)

# See https://github.com/color-js/color.js/blob/a77e080a070039c534dda3965a769675aac5f75e/src/spaces/p3-linear.js

_XYZ_TO_LINEAR_P3 = (
	(  2.493496911941425,   -0.9313836179191239,  -0.40271078445071684  ),
	( -0.8294889695615747,   1.7626640603183463,   0.023624685841943577 ),
	(  0.03584583024378447, -0.07617238926804182,  0.9568845240076872   ),
)

_LINEAR_P3_TO_XYZ = (
	( 0.4865709486482162, 0.26566769316909306, 0.1982172852343625 ),
	( 0.2289745640697488, 0.6917385218365064,  0.079286914093745  ),
	( 0.0000000000000000, 0.04511338185890264, 1.043944368900976  ),
)

# See https://github.com/color-js/color.js/blob/a77e080a070039c534dda3965a769675aac5f75e/src/spaces/oklab.js

_XYZ_TO_LMS = (
	( 0.8190224379967030, 0.3619062600528904, -0.1288737815209879 ),
	( 0.0329836539323885, 0.9292868615863434,  0.0361446663506424 ),
	( 0.0481771893596242, 0.2642395317527308,  0.6335478284694309 ),
)

_LMS_TO_XYZ = (
	(  1.2268798758459243, -0.5578149944602171,  0.2813910456659647 ),
	( -0.0405757452148008,  1.1122868032803170, -0.0717110580655164 ),
	( -0.0763729366746601, -0.4214933324022432,  1.5869240198367816 ),
)

_LMS_TO_OKLAB = (
	( 0.2104542683093140,  0.7936177747023054, -0.0040720430116193 ),
	( 1.9779985324311684, -2.4285922420485799,  0.4505937096174110 ),
	( 0.0259040424655478,  0.7827717124575296, -0.8086757549230774 ),
)

_OKLAB_TO_LMS = (
    ( 1.0000000000000000,  0.3963377773761749,  0.2158037573099136 ),
	( 1.0000000000000000, -0.1055613458156586, -0.0638541728258133 ),
	( 1.0000000000000000, -0.0894841775298119, -1.2914855480194092 ),
)


# --------------------------------------------------------------------------------------


_Vector: TypeAlias = tuple[float, float, float]
_Matrix: TypeAlias = tuple[_Vector, _Vector, _Vector]

def _multiply(matrix: _Matrix, vector: _Vector) -> _Vector:
    return cast(
        _Vector,
        tuple(sum(r * c for r, c in zip(row, vector)) for row in matrix)
    )


# --------------------------------------------------------------------------------------
# 24-bit RGB


def rgb256_to_srgb(r: int, g: int, b: int) -> tuple[float, float, float]:
    """Convert the given color from 24-bit RGB to sRGB."""
    return cast(_Vector, tuple(map(lambda c: c / 255.0, (r, g, b))))


def srgb_to_rgb256(r: float, g: float, b: float) -> tuple[int, int, int]:
    """
    :bdg-warning:`Lossy conversion` Convert the given color from sRGB to 24-bit
    RGB.
    """
    return cast(tuple[int, int, int], tuple(map(lambda c: round(c * 255), (r, g, b))))


# --------------------------------------------------------------------------------------
# sRGB and Linear sRGB
# See https://github.com/color-js/color.js/blob/a77e080a070039c534dda3965a769675aac5f75e/src/spaces/srgb.js


def srgb_to_linear_srgb(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the given color from sRGB to linear sRGB."""
    def convert(value: float) -> float:
        magnitude = math.fabs(value)

        if magnitude <= 0.04045:
            return value / 12.92

        return math.copysign(math.pow((magnitude + 0.055) / 1.055, 2.4), value)

    return convert(r), convert(g), convert(b)


def linear_srgb_to_srgb(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the given color from linear sRGB to sRGB."""
    def convert(value: float) -> float:
        magnitude = math.fabs(value)

        if magnitude <= 0.0031308:
            return value * 12.92

        return math.copysign(math.pow(magnitude, 1/2.4) * 1.055 - 0.055, value)

    return convert(r), convert(g), convert(b)


def linear_srgb_to_xyz(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the given color from linear sRGB to XYZ."""
    return _multiply(_LINEAR_SRGB_TO_XYZ, (r, g, b))


# --------------------------------------------------------------------------------------
# P3 and Linear P3


p3_to_linear_p3 = srgb_to_linear_srgb
linear_p3_to_p3 = linear_srgb_to_srgb


def linear_p3_to_xyz(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the given color from linear P3 to XYZ."""
    return _multiply(_LINEAR_P3_TO_XYZ, (r, g, b))


# --------------------------------------------------------------------------------------
# Oklab and Oklch:
# See https://github.com/color-js/color.js/blob/a77e080a070039c534dda3965a769675aac5f75e/src/spaces/oklch.js


def oklch_to_oklab(L: float, C: float, h: float) -> tuple[float, float, float]:
    """Convert the given color from Oklch to Oklab."""
    if math.isnan(h):
        a = b = 0.0
    else:
        a = C * math.cos(h * math.pi / 180)
        b = C * math.sin(h * math.pi / 180)

    return L, a, b


def oklab_to_oklch(L: float, a: float, b: float) -> tuple[float, float, float]:
    """Convert the given color from Oklab to Oklch."""
    ε = 0.0002

    if math.fabs(a) < ε and math.fabs(b) < ε:
        h = math.nan
    else:
        h = math.atan2(b, a) * 180 / math.pi

    return L, math.sqrt(math.pow(a, 2) + math.pow(b, 2)), math.fmod(h + 360, 360)


def oklab_to_xyz(L: float, a: float, b: float) -> tuple[float, float, float]:
    """Convert the given color from Oklab to XYZ."""
    LMSg = _multiply(_OKLAB_TO_LMS, (L, a, b))
    LMS = cast(_Vector, tuple(map(lambda c: math.pow(c, 3), LMSg)))
    return _multiply(_LMS_TO_XYZ, LMS)


# --------------------------------------------------------------------------------------
# XYZ


def xyz_to_linear_srgb(X: float, Y: float, Z: float) -> tuple[float, float, float]:
    """Convert the given color from XYZ to linear sRGB."""
    return _multiply(_XYZ_TO_LINEAR_SRGB, (X, Y, Z))


def xyz_to_linear_p3(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the given color from XYZ to linear P3."""
    return _multiply(_XYZ_TO_LINEAR_P3, (r, g, b))


def xyz_to_oklab(X: float, Y: float, Z: float) -> tuple[float, float, float]:
    """Convert the given color from XYZ to Oklab."""
    LMS = _multiply(_XYZ_TO_LMS, (X, Y, Z))
    LMSg = cast(_Vector, tuple(map(lambda c: math.cbrt(c), LMS)))
    return _multiply(_LMS_TO_OKLAB, LMSg)


# --------------------------------------------------------------------------------------
# Arbitrary Conversions


def _collect_conversions(
    mod: dict[str, object],
    conversions: dict[str, dict[str, ConverterSpec]]
) -> None:
    for name, value in mod.items():
        if not name.startswith('_') and '_to_' in name and callable(value):
            source, _, target = name.partition('_to_')
            targets = conversions.setdefault(source, {})
            if target in targets:
                raise ValueError(f'duplicate conversion from {source} to {target}')
            targets[target] = value


_BASE_TREE = {
    'rgb256': ('srgb', 3),
    'srgb': ('linear_srgb', 2),
    'linear_srgb': ('xyz', 1),
    'p3': ('linear_p3', 2),
    'linear_p3': ('xyz', 1),
    'oklch': ('oklab', 2),
    'oklab': ('xyz', 1),
    'xyz': (None, 0),
}

def _elaborate_route(source: str, target: str) -> tuple[str, ...]:
    """Elaborate the route from the source to the target color format or space."""
    if source not in _BASE_TREE:
        raise ValueError(f'{source} is not a valid color format or space')
    if target not in _BASE_TREE:
        raise ValueError(f'{target} is not a valid color format or space')

    # Trace paths from source and target towards root of base tree
    source_path: list[str] = [source]
    target_path: list[str] = [target]

    def step(path: list[str]) -> bool:
        tag, _ = _BASE_TREE[path[-1]]
        if tag is not None:
            path.append(tag)
        return tag is None

    # Sync up traces, so that both have same distance from root
    _, source_dist = _BASE_TREE[source]
    _, target_dist = _BASE_TREE[target]

    path = source_path if source_dist >= target_dist else target_path
    for _ in range(abs(source_dist - target_dist)):
        done = step(path)
        assert not done

    # Keep tracing in lock step until paths share last node
    while source_path[-1] != target_path[-1]:
        done = step(source_path)
        done |= step(target_path)
        assert not done

    # Assemble complete path
    target_path.pop()
    target_path.reverse()
    return tuple(itertools.chain(source_path, target_path))


def _pass_through(*coordinates: float) -> CoordinateSpec:
    """Pass through the coordinates."""
    return cast(CoordinateSpec, tuple(coordinates))


def _create_converter(conversions: tuple[ConverterSpec, ...]) -> ConverterSpec:
    """
    Instantiate a closure that applies the given conversions. Doing so in a
    dedicated top-level function keeps the closure environment minimal.
    """
    def converter(*coordinates: float) -> CoordinateSpec:
        value = cast(CoordinateSpec, coordinates)
        for fn in conversions:
            value = fn(*value)  # type: ignore
        return value
    return cast(ConverterSpec, converter)


_LORES_FORMAT = {'ansi', 'eight_bit', 'rgb6'}

_converter_cache: dict[str, dict[str, ConverterSpec]] = {}
_collected_mod_lores: bool = False

def get_converter(source: str, target: str) -> ConverterSpec:
    """
    Instantiate a function that converts coordinates from the source color
    format or space to the target color format or space.

    This function factory caches converters to avoid re-instantiating the same
    converter over and over again. Each converter's name is computed as
    ``f"{source}_to_{target}"``.
    """
    # Initialize converter cache with basic conversions
    global _collected_mod_lores

    if not _converter_cache:
        _collect_conversions(globals(), _converter_cache)

        for tag in _LORES_FORMAT:
            _converter_cache.setdefault(tag, {})

    # Handle trivial case
    if source == target:
        return cast(ConverterSpec, _pass_through)

    # Check whether converter already exists
    maybe_converter = _converter_cache[source].get(target)
    if maybe_converter is not None:
        return maybe_converter

    # If either format is lo-res, make sure lo-res conversions have been loaded
    is_source_lores = source in _LORES_FORMAT
    is_target_lores = target in _LORES_FORMAT
    if (is_source_lores or is_target_lores) and not _collected_mod_lores:
        pkg, _, _ = __name__.rpartition('.')
        from importlib import import_module
        mod = import_module('.lores', pkg)
        _collect_conversions(vars(mod), _converter_cache)
        _collected_mod_lores = True

    # Determine route. Only the first and/or last node can be lo-res.
    route: list[str] = [source] if is_source_lores else []

    route.extend(_elaborate_route(
        'rgb256' if is_source_lores else source,
        'oklab' if is_target_lores else target,
    ))

    if is_target_lores:
        route.append(target)

    # Turn list of nodes into list of functions into converter function
    conversions = tuple(
        _converter_cache[t1][t2] for t1, t2 in itertools.pairwise(route)
    )

    # Annotate converter for easy debugability
    converter = _create_converter(conversions)
    name = f'{source}_to_{target}'
    setattr(converter, '__name__', name)
    setattr(converter, '__qualname__', name)
    setattr(converter, 'route', route)
    setattr(converter, 'conversions', conversions)

    _converter_cache[source][target] = converter
    return converter



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'prettypretty'
copyright = '© 2024 Robert Grimm'
author = 'Robert Grimm'

# Source order is nicer than alphabetizing fields and methods.
autodoc_member_order = 'bysource'

# Known types are linked, so the fully qualified name adds mostly noise.
python_use_unqualified_type_names = True

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    #"sphinx.ext.intersphinx",
    #"sphinx.ext.extlinks",
    "sphinx.ext.viewcode",
    "sphinxcontrib.autoprogram",
    "sphinx_design",
    "sphinx_rtd_theme",
]

napoleon_include_init_with_doc = True
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

#html_theme = "sphinxawesome"
html_static_path = ['_static']
html_theme = 'sphinx_rtd_theme'



"""Support for serializing and deserializing color values"""
import enum
from typing import cast, Literal, NoReturn, overload


@overload
def _check(
    is_valid: Literal[False], entity: str, value: object, deficiency: str = ...
) -> NoReturn:
    ...
@overload
def _check(
    is_valid: bool, entity: str, value: object, deficiency: str = ...
) -> None | NoReturn:
    ...
def _check(
    is_valid: bool, entity: str, value: object, deficiency: str = 'is malformed'
) -> None | NoReturn:
    if not is_valid:
        raise SyntaxError(f'{entity} "{value}" {deficiency}')
    return


def parse_hex(color: str) -> tuple[str, tuple[int, int, int]]:
    """Parse the string specifying a color in hashed hexadecimal format."""
    entity = 'hex web color'

    try:
        _check(color.startswith('#'), entity, color, 'does not start with "#"')
        color = color[1:]
        digits = len(color)
        _check(digits in (3, 6), entity, color, 'does not have 3 or 6 digits')
        if digits == 3:
            color = ''.join(f'{d}{d}' for d in color)
        return 'rgb256', cast(
            tuple[int, int, int],
            tuple(int(color[n:n+2], base=16) for n in range(0, 6, 2)),
        )
    except SyntaxError:
        raise
    except:
        _check(False, entity, color)


def parse_x_rgb(color: str) -> tuple[str, tuple[float, float, float]]:
    """Parse the string specifying a color in X's rgb: format."""
    entity = 'X rgb color'

    try:
        _check(color.startswith('rgb:'), entity, color, 'does not start with "rgb:"')
        hexes = [(f'{h}{h}' if len(h) == 1 else h) for h in color[4:].split('/')]
        _check(len(hexes) == 3, entity, color, 'does not have three components')
        if max(len(h) for h in hexes) == 2:
            return 'rgb256', cast(
                tuple[int, int, int],
                tuple(int(h, base=16) for h in hexes)
            )
        else:
            return 'srgb', tuple(int(h, base=16) / 16 ** len(h) for h in hexes)
    except SyntaxError:
        raise
    except:
        _check(False, entity, color)


def parse_x_rgbi(color: str) -> tuple[str, tuple[float, float, float]]:
    """Parse the string specifying a color in X's rgbi: format."""
    entity = 'X rgbi color'

    try:
        _check(color.startswith('rgbi:'), entity, color, 'does not start with "rgbi:"')
        cs = [float(c) for c in color[5:].split('/')]
        _check(len(cs) == 3, entity, color, 'does not have three components')
        for c in cs:
            _check(0 <= c <= 1, entity, color, 'has non-normal component')
        return 'srgb', cast(tuple[float, float, float], tuple(cs))
    except SyntaxError:
        raise
    except:
        _check(False, entity, color)


class Format(enum.Enum):
    """
    The color format

    Attributes:
        FUNCTION: for ``<tag>(<coordinates>)`` notation
        HEX: for ``#<hex>`` notation
        CSS: for ``color()``, ``oklab()``, ``oklch()``, and ``rgb()`` notation
        X: for ``rgb:<hex>/<hex>/<hex>`` and ``rgbi:<float>/<float>/<float>``
            notation
    """
    FUNCTION = 'f'
    HEX = 'h'
    CSS = 's'
    X = 'x'


def parse_format_spec(spec: str) -> tuple[Format, int]:
    """
    Parse the color format specifier into the format and precision.

    Args:
        spec: selects the desired output format and precision
    Returns:
        the format and maximum precision for floating point numbers, which
        default to `Format.FUNCTION` and 5, respectively

    A valid format specifier comprises two parts, both of which are optional:

     1. The first part, if present specifies the precision and is written as a
        period followed by one or two decimal digits, e.g., ``.3``.
     2. The second part, if present, specifies the format:

          * ``f`` for function notation, which uses the tag as function name
            and the comma-separated coordinates as arguments
          * ``h`` for hexadecimal notation prefixed with a hash ``#``
          * ``s`` for CSS notation, which uses CSS function and color space
            names and space-separated coordinates
          * ``x`` for X notation, which uses ``rgb:`` or ``rgbi:`` as a prefix

    Note that the terminal-specific ``ansi``, ``eight_bit``, and ``rgb6`` color
    formats as well as the ``linear_p3`` color space have no CSS serialization.
    Also note that ``h`` only works for RGB256 colors and ``x`` only for RGB256
    and sRGB colors.
    """
    format = Format.FUNCTION
    precision = 5

    s = spec
    if s:
        f = s[-1]
        if f in ('f', 'h', 's', 'x'):
            format = Format(f)
            s = s[:-1]
    if s.startswith('.'):
        precision = int(s[1:])
        s = ''
    if s:
        raise ValueError(f'malformed color format "{spec}"')

    return format, precision


_CSS_FORMATS = {
    'rgb256': 'rgb({})',
    'srgb': 'color(srgb {})',
    'linear_srgb': 'color(srgb-linear {})',
    'p3': 'color(display-p3 {})',
    'xyz': 'color(xyz {})',
    'oklab': 'oklab({})',
    'oklch': 'oklch({})',
}

def stringify(
    tag: str,
    coordinates: tuple[int] | tuple[float, float, float],
    format: Format = Format.FUNCTION,
    precision: int = 5
) -> str:
    """
    Format the tagged coordinates in the specified format and with the specified
    precision.
    """
    if format is Format.HEX:
        return '#' + ''.join(f'{c:02x}' for c in coordinates)
    elif format is Format.X:
        if all(isinstance(c, int) for c in coordinates):
            return 'rgb:' + '/'.join(f'{c:02x}' for c in coordinates)
        else:
            return 'rgbi:' + '/'.join(f'{float(c):.{precision}}' for c in coordinates)

    separator = ' ' if format is Format.CSS else ', '
    coordinate_text = separator.join(
        f'{c}' if isinstance(c, int) else f'{c:.{precision}}'
        for c in coordinates
    )

    if format is Format.FUNCTION:
        return f'{tag}({coordinate_text})'

    css_format = _CSS_FORMATS.get(tag)
    if css_format is None:
        raise ValueError(f'{tag} has no CSS serialization')
    return css_format.format(coordinate_text)



import enum


class Layer(enum.Enum):
    """The two layers for terminal presentation, text and background."""
    TEXT = 0
    BACKGROUND = 10


def eight_bit_to_sgr_params(color: int, layer: Layer) -> tuple[int, ...]:
    """Convert the 8-bit terminal color to the corresponding SGR parameters."""
    if 0 <= color <= 7:
        return 30 + layer.value + color,
    if 8 <= color <= 15:
        return 90 + layer.value + color - 8,
    return 38 + layer.value, 5, color


def sgr(*parameters: int) -> str:
    """Create an SGR escape sequence for the given parameters."""
    return f'\x1b[{";".join(str(p) for p in parameters)}m'



"""
A script to visualize 8-bit terminal colors as well as prettypretty's support
for down-sampling colors and maximizing contrast.
"""
import argparse
import os

from .conversion import rgb256_to_srgb, get_converter
from .apca import use_black_text, use_black_background
from .lores import (
    rgb6_to_eight_bit,
    rgb6_to_rgb256,
    eight_bit_to_rgb256,
    oklab_to_ansi,
    oklab_to_eight_bit,
)
from .theme import MACOS_TERMINAL, VGA, XTERM, current_theme
from .style import eight_bit_to_sgr_params, Layer, sgr

class FramedBoxes:
    """
    Boxes in a frame.

    This class helps with incrementally formatting a grid of boxes with a
    surrounding frame (or border). Each box has the same width and a height of
    one line. Boxes are formatted left-to-right.
    """
    width: int
    box_count: int
    box_width: int
    fragments: list[str]

    def __init__(self, width: int, box_count: int = 1, min_width: int = 5) -> None:
        self._box_count = box_count
        self._box_width = (width - 2) // box_count
        if self._box_width < min_width:
            raise ValueError(f'unable to fit {box_count} boxes into {width} columns')
        self._width = self._box_count * self._box_width + 2
        self._fragments: list[str] = []
        self._line_content_width = 0

    @property
    def outer_width(self) -> int:
        """The total width of the frame."""
        return self._width

    @property
    def inner_width(self) -> int:
        """
        The width of the content inside the frame, which is ``outer_width - 2``.
        """
        return self._width - 2

    def top(self, title: str = '') -> None:
        """Format the top of the frame."""
        title = f' {t} ' if (t := title.strip()) != '' else t
        filling = self.outer_width - 4 - len(title)
        if filling < 0:
            raise ValueError(f'"{title}" is too long for {self.outer_width}-wide frame')
        if title:
            title = f'\x1b[1m{title}\x1b[m'
        self._fragments.append(f'┏━{title}{"━" * filling}━┓\n')

    def left(self) -> None:
        """Start formatting a line of content."""
        self._fragments.append('┃')
        if self._line_content_width != 0:
            raise ValueError('Line started before line ended')

    def box(
        self,
        text: str,
        foreground: tuple[int] | tuple[int, int, int],
        background: tuple[int] | tuple[int, int, int]
    ) -> None:
        """Format one box of the content."""
        box = text.center(self._box_width)
        if len(box) != self._box_width:
            raise ValueError(f'"{text}" does not fit into {self._box_width}-wide box')

        if len(background) == 1:
            bg_params = eight_bit_to_sgr_params(*background, Layer.BACKGROUND)
        else:
            bg_params = 48, 2, *background

        if len(foreground) == 1:
            fg_params = eight_bit_to_sgr_params(*foreground, Layer.TEXT)
        else:
            fg_params = 38, 2, *foreground

        self._fragments.append(f'{sgr(*bg_params, *fg_params)}{box}\x1b[m')
        self._line_content_width += self._box_width

    def right(self) -> None:
        """Complete formatting a line of content."""
        if self._line_content_width != self.inner_width:
            raise ValueError(
                f'content spans {self._line_content_width}, '
                f'not {self.inner_width} columns'
            )

        self._fragments.append('┃\n')
        self._line_content_width = 0

    def bottom(self) -> None:
        """Format the bottom of the frame."""
        self._fragments.append(f'┗{"━" * self.inner_width}┛')

    def __str__(self) -> str:
        return ''.join(self._fragments)


def format_color_cube(
    width: int,
    *,
    layer: Layer = Layer.BACKGROUND,
    ansi_only: bool = False,
) -> str:
    """
    Format a framed grid with 216 cells, where each cell displays a distinct
    color from the 6x6x6 cube of 8-bit terminal colors.

    Args:
        width: is the number of columns available to the framed grid
        layer: determines whether to color text or background, with background
            the default
        ansi_only: determines whether to down-sample the 8-bit color to an
            extended ANSI color

    Returns:
        The fully formated and framed grid
    """
    frame = FramedBoxes(width, 6)
    frame.top(
        layer.name.capitalize()
        + ': '
        + ('Downsampled ' if ansi_only else '')
        + '6•6•6 RGB Cube'
    )

    rgb256_to_oklab = get_converter('rgb256', 'oklab')

    for r in range(6):
        for b in range(6):
            frame.left()

            for g in range(6):
                rgb256 = rgb6_to_rgb256(r, g, b)

                if ansi_only:
                    eight_bit = oklab_to_ansi(*rgb256_to_oklab(*rgb256))
                    srgb = rgb256_to_srgb(*eight_bit_to_rgb256(*eight_bit))
                else:
                    eight_bit = rgb6_to_eight_bit(r, g, b)
                    srgb = rgb256_to_srgb(*rgb256)

                # Pick black or white for other color based on contrast
                if layer is Layer.BACKGROUND:
                    foreground = (232 if use_black_text(*srgb) else 255),
                    background = eight_bit
                else:
                    foreground = eight_bit
                    background = (232 if use_black_background(*srgb) else 255),

                frame.box(f'{r}•{g}•{b}', foreground, background)

            frame.right()

    frame.bottom()
    return str(frame)


def format_hires_slice(
    width: int,
    *,
    eight_bit_only: bool = False,
) -> str:
    frame = FramedBoxes(width, 32, min_width=1)
    frame.top(
        'High-Resolution Color Slice'
    )

    rgb256_to_oklab = get_converter('rgb256', 'oklab')

    for r in range(0, 256, 8):
        frame.left()
        g = 0

        for b in range(0, 256, 8):
            color = r, g, b
            if eight_bit_only:
                color = oklab_to_eight_bit(*rgb256_to_oklab(*color))
            frame.box(' ', (0,), color)

        frame.right()

    frame.bottom()
    return str(frame)


def create_parser() -> argparse.ArgumentParser:
    """Create a command line argument parser."""
    parser = argparse.ArgumentParser(
        description="""
            Display color grids that visualize the range of terminal colors,
            while also exercising prettypretty's support for maximizing
            contrast and down-sampling colors.
        """,
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--macos-terminal',
        action='store_const',
        const=MACOS_TERMINAL,
        dest='theme',
        help="use the same colors as the Basic theme for macOS Terminal"
    )
    group.add_argument(
        '--xterm',
        action='store_const',
        const=XTERM,
        dest='theme',
        help='use the same colors as xterm'
    )
    group.add_argument(
        '--vga',
        action='store_const',
        const=VGA,
        dest='theme',
        help='use the same colors as VGA in text mode'
    )

    return parser


if __name__ == '__main__':
    options = create_parser().parse_args()
    width, _ = os.get_terminal_size()

    with current_theme(options.theme or VGA):
        print(f'\n{format_color_cube(width)}')
        print(f'\n{format_color_cube(width, ansi_only=True)}')
        print(f'\n{format_color_cube(width, layer=Layer.TEXT)}')

        if os.getenv('COLORTERM') == 'truecolor':
            print(f'\n{format_hires_slice(width)}')
            print(f'\n{format_hires_slice(width, eight_bit_only=True)}')



"""
Basic type declarations for coordinates, colors, and conversions:

  * ``IntCoordinateSpec`` and ``FloatCoordinateSpec`` are triples of integers
    and floating point values, respectively
  * ``CoordinateSpec`` combines the two types with a tuple of a single integer
  * ``ConverterSpec`` describes a function that converts from one color format
    or space into another
  * ``ColorSpec`` is a dataclass that associates a color's components with
    the tag identifying the color format or space.

All container types are immutable.
"""
import dataclasses
from typing import overload, Protocol, TypeAlias

IntCoordinateSpec: TypeAlias = tuple[int, int, int]
FloatCoordinateSpec: TypeAlias = tuple[float, float, float]
CoordinateSpec: TypeAlias = (
    tuple[int] | tuple[int, int, int] | tuple[float, float, float]
)

CoordinateVectorSpec: TypeAlias = tuple[FloatCoordinateSpec, ...]

class ConverterSpec(Protocol):
    @overload
    def __call__(
        self, __c: int
    ) -> tuple[int] | tuple[int, int, int]: ...
    @overload
    def __call__(
        self, __c1: int, __c2: int, __c3: int
    ) -> tuple[int, int, int]: ...
    @overload
    def __call__(
        self, __c1: float, __c2: float, __c3: float
    ) -> tuple[float, float, float]: ...

    def __call__(
        self,
        __c1: int | float,
        __c2: None | int | float = None,
        __c3: None | int | float = None,
    ) -> CoordinateSpec:
        ...


_TAGS = {
    'ansi': '1i',
    'eight_bit': '1i',
    'linear_p3': '3f',
    'linear_srgb': '3f',
    'oklab': '3f',
    'oklch': '3f',
    'p3': '3f',
    'rgb6': '3i',
    'rgb256': '3i',
    'srgb': '3f',
    'xyz': '3f',
}


@dataclasses.dataclass(frozen=True, slots=True)
class ColorSpec:
    """
    A color specification.

    Attributes:
        tag: identifies the coordinates' color format or space
        coordinates: are the numeric components of the color

    With exception of the ``ansi`` and ``eight_bit`` terminal color formats,
    which have one components, all other color formats and spaces have three
    components. In particular, the ``rgb6`` and ``rgb256`` formats have three
    integer components, whereas the color spaces all have three floating point
    components.

    This class does validate the tag and number of coordinates upon creation.

    Instance of this class are immutable.
    """
    tag: str
    coordinates: CoordinateSpec

    def __post_init__(self) -> None:
        code = _TAGS.get(self.tag)
        if code is None:
            raise ValueError(f'{self.tag} is not a valid color format or space')
        count = 1 if code == '1i' else 3
        if (l := len(self.coordinates)) != count:
            raise ValueError(f'{self.tag} should have {count} coordinates, not {l}')



"""Support for computing the difference between two or more colors"""
from collections.abc import Iterable
import math
from typing import Literal


def deltaE_oklab(
    L1: float, a1: float, b1: float,
    L2: float, a2: float, b2: float,
    *,
    version: Literal[1, 2] = 1,
) -> float:
    """
    Determine the difference between two Oklab colors.

    For the first version, that difference is just the Euclidian distance
    between the coordinates.

    For the second version, the difference between the *a* and *b* coordinates
    is scaled by a constant factor before computing the Euclidian distance. For
    now that factor is 2, even though `it probably is closer to 2.1
    <https://github.com/w3c/csswg-drafts/issues/6642#issuecomment-945714988>`_.
    """
    ΔL = L1 - L2
    Δa = version * (a1 - a2)
    Δb = version * (b1 - b2)
    return math.sqrt(ΔL * ΔL + Δa * Δa + Δb * Δb)


def closest_oklab(
    origin: tuple[float, float, float],
    candidates: Iterable[tuple[float, float, float]],
) -> tuple[int, tuple[float, float, float]]:
    """
    Find the color closest to the origin amongst candidate colors.

    Args:
        origin: is the reference color in Oklab coordinates
        candidates: are the colors to compare to, also in Oklab coordinates
    Returns:
        the index and coordinates of the candidate color closest to the origin,
        which are -1 and ``origin`` if the iterable is empty.

    This function iterates over the candidates only once and hence the iterable
    may also be an iterator.
    """
    min_ΔE = math.inf
    min_index = -1
    min_color = origin

    for index, color in enumerate(candidates):
        ΔE = deltaE_oklab(*origin, *color)
        if ΔE < min_ΔE:
            min_ΔE = ΔE
            min_index = index
            min_color = color

    return min_index, min_color



"""
Support for computing color contrast with the `accessible perceptual contrast
algorithm <https://github.com/Myndex/apca-w3>`_ (APCA), version 0.1.9, dated
July 3, 2022. Note that APCA is asymmetric, i.e., the order of text and
background luminance matters. Furthermore, it is designed for text on a solidly
colored background only and clamps small contrast values to zero.
"""
import math


_APCA_EXPONENT = 2.4
_APCA_COEFFICIENTS = (0.2126729, 0.7151522, 0.0721750)

_APCA_BLACK_THRESHOLD = 0.022
_APCA_BLACK_CLIP = 1.414
_APCA_DELTA_Y_MIN = 0.0005

_APCA_BoW_TEXT = 0.57
_APCA_BoW_BACKGROUND = 0.56
_APCA_WoB_TEXT = 0.62
_APCA_WoB_BACKGROUND = 0.65

_APCA_SCALE = 1.14
_APCA_CLAMP = 0.1
_APCA_OFFSET = 0.027


def srgb_to_luminance(r: float, g: float, b: float) -> float:
    """
    :bdg-warning:`Internal API` Determine the non-standard APCA luminance for
    the given sRGB color.
    """
    linear_srgb = (math.pow(c, _APCA_EXPONENT) for c in (r, g, b))
    return sum(c1 * c2 for c1, c2 in zip(_APCA_COEFFICIENTS, linear_srgb))


def luminance_to_contrast(
    text_luminance: float,
    background_luminance: float,
) -> float:
    """
    :bdg-warning:`Internal API` Determine the contrast between the text and
    background luminance values."""
    assert 0.0 <= text_luminance <= 1.1 and 0.0 <= background_luminance <= 1.1

    # Soft-clip and clamp black
    if text_luminance < _APCA_BLACK_THRESHOLD:
        text_luminance += math.pow(
            _APCA_BLACK_THRESHOLD - text_luminance,
            _APCA_BLACK_CLIP,
        )
    if background_luminance < _APCA_BLACK_THRESHOLD:
        background_luminance += math.pow(
            _APCA_BLACK_THRESHOLD - background_luminance,
            _APCA_BLACK_CLIP,
        )

    # Small ΔY have too little contrast. Clamp result to zero.
    if abs(text_luminance - background_luminance) < _APCA_DELTA_Y_MIN:
        return 0.0

    # Computer Lc
    if background_luminance > text_luminance:
        contrast = (
            math.pow(background_luminance, _APCA_BoW_BACKGROUND)
            - math.pow(text_luminance, _APCA_BoW_TEXT)
        ) * _APCA_SCALE

        if contrast < _APCA_CLAMP:
            return 0.0
        return contrast - _APCA_OFFSET

    else:
        contrast = (
            math.pow(background_luminance, _APCA_WoB_BACKGROUND)
            - math.pow(text_luminance, _APCA_WoB_TEXT)
        ) * _APCA_SCALE

        if contrast > -_APCA_CLAMP:
            return 0.0
        return contrast + _APCA_OFFSET


def contrast(
    text_color: tuple[float, float, float],
    background_color: tuple[float, float, float],
) -> float:
    """
    Compute the contrast between the given text and background colors in the
    sRGB color space.
    """
    text_luminance = srgb_to_luminance(*text_color)
    background_luminance = srgb_to_luminance(*background_color)
    return luminance_to_contrast(text_luminance, background_luminance)


def use_black_text(r: float, g: float, b: float) -> bool:
    """
    Determine whether text should be black or white to maximize its contrast
    against a background with the given sRGB color.
    """
    luminance = srgb_to_luminance(r, g, b)

    return (
        luminance_to_contrast(0.0, luminance) >= -luminance_to_contrast(1.0, luminance)
    )


def use_black_background(r: float, g: float, b: float) -> bool:
    """
    Determine whether the background should be black or white to maximize
    its contrast against text with the given sRGB color.
    """
    luminance = srgb_to_luminance(r, g, b)

    return (
        luminance_to_contrast(luminance, 0.0) <= -luminance_to_contrast(luminance, 1.0)
    )



"""
Support for low-resolution terminal colors
"""
from itertools import chain
from typing import Callable, cast

# Rename basic conversion to prevent it from tripping up module scan
from .conversion import get_converter, rgb256_to_srgb as rgb256_into_srgb
from .difference import closest_oklab
from .spec import CoordinateVectorSpec
from .theme import current_theme, Theme


_RGB6_TO_RGB256 = (0, 0x5F, 0x87, 0xAF, 0xD7, 0xFF)


def rgb6_to_eight_bit(r: int, g: int, b: int) -> tuple[int]:
    """
    Convert the given color from the 6x6x6 RGB cube of 8-bit terminal colors to
    an actual 8-bit terminal color.
    """
    assert 0 <= r <= 5 and 0 <= g <= 5 and 0 <= b <= 5
    return 16 + 36 * r + 6 * g + b,


def eight_bit_to_rgb6(color: int) -> tuple[int, int, int]:
    """
    Convert the given 8-bit color to the three components of the 6x6x6 RGB cube.
    The color value must be between 16 and 231, inclusive.
    """
    assert 16 <= color <= 231

    b = color - 16
    r = b // 36
    b -= 36 * r
    g = b // 6
    b -= 6 * g
    return r, g, b


def rgb6_to_rgb256(r: int, g: int, b: int) -> tuple[int, int, int]:
    """Convert the given color in RGB6 format to RGB256 format."""
    assert 0 <= r <= 5 and 0 <= g <= 5 and 0 <= b <= 5
    return _RGB6_TO_RGB256[r], _RGB6_TO_RGB256[g], _RGB6_TO_RGB256[b]


def approximate_rgb256_with_rgb6(r: int, g: int, b: int) -> tuple[int, int, int]:
    """
    :bdg-warning:`Lossy conversion` Convert the given color from RGB256 to RGB6.

    This function effectively reverses the conversion from RGB6 to RGB256: It
    compares each RGB256 coordinate with the RGB256 values used for the inverse
    and picks the RGB6 with the closest RGB256 value for the inverse.

    The correctness of this particular implementation depends on the inverse
    mapping the extrema of the domain to the extrema of the codomain, i.e., 0 to
    0 and 5 to 255.
    """
    assert 0 <= r <= 255 and 0 <= g <= 255 and 0 <= b <= 255

    def convert(value: int) -> int:
        for index, level in enumerate(_RGB6_TO_RGB256):
            if value == level:
                return index
            if value > level:
                continue

            # The RGB256 value is between two RGB6 values. Pick the closer one.
            previous_level = _RGB6_TO_RGB256[index - 1]
            return index if level - value < value - previous_level else index - 1

        assert False, 'unreachable statement'

    return convert(r), convert(g), convert(b)


def ansi_to_eight_bit(color: int) -> tuple[int]:
    """Convert the given ANSI color to 8-bit format."""
    return color,


def ansi_to_rgb256(color: int) -> tuple[int, int, int]:
    """
    :bdg-warning:`Lossy conversion` Convert the given ANSI color to RGB256
    format.

    .. warning::
        The result of this function critically depends on the current color
        theme. After all, the current theme determines the RGB256 values for all
        extended ANSI colors.
    """
    assert 0 <= color <= 15
    c = current_theme().ansi(color)

    assert c.tag == 'rgb256'
    return c.coordinates  # type: ignore


def ansi_to_srgb(color: int) -> tuple[float, float, float]:
    """
    Convert the ANSI color to sRGB. Directly converting to sRGB and avoiding
    RGB256 is the more conservative conversion because most terminals, when
    queried with OSC-4, report color values with four hexadecimal digits per
    coordinate. RGB256 obviously cannot preserve that resolution, though sRGB
    can.

    .. warning::
        The result of this function critically depends on the current color
        theme. After all, the current theme determines the RGB256 values for all
        extended ANSI colors.
    """
    assert 0 <= color <= 15
    c = current_theme().ansi(color)

    if c.tag == 'rg256':
        coordinates = rgb256_into_srgb(*cast(tuple[int, int, int], c.coordinates))
    elif c.tag == 'srgb':
        coordinates = c.coordinates
    else:
        coordinates = get_converter(c.tag, 'srgb')(*c.coordinates)

    return cast(tuple[float, float, float], coordinates)


def _eight_bit_gray_to_rgb256(color: int) -> tuple[int, int, int]:
    """Convert the given 8-bit gray to RGB256 format."""
    assert 232 <= color <= 255
    c = 10 * (color - 232) + 8
    return c, c, c


def eight_bit_to_rgb256(color: int) -> tuple[int, int, int]:
    """
    :bdg-warning:`Lossy conversion` Convert the given 8-bit terminal color to
    24-bit RGB.

    .. warning::
        The result of this function may depend on the current color theme.
        It provides RGB256 color values for 8-bit colors 0–15, i.e., the
        extended ANSI colors.
    """
    if 0 <= color <= 15:
        return ansi_to_rgb256(color)
    if 16 <= color <= 231:
        return rgb6_to_rgb256(*eight_bit_to_rgb6(color))
    if 232 <= color <= 255:
        return _eight_bit_gray_to_rgb256(color)

    raise ValueError(f'{color} is not a valid 8-bit terminal color')


def eight_bit_to_srgb(color: int) -> tuple[float, float, float]:
    """
    Convert the given 8-bit terminal color to sRGB.

    .. warning::
        The result of this function may depend on the current color theme.
        It provides RGB256 color values for 8-bit colors 0–15, i.e., the
        extended ANSI colors.
    """
    if 0 <= color <= 15:
        return ansi_to_srgb(color)
    if 16 <= color <= 231:
        return rgb256_into_srgb(*rgb6_to_rgb256(*eight_bit_to_rgb6(color)))
    if 232 <= color <= 255:
        return rgb256_into_srgb(*_eight_bit_gray_to_rgb256(color))

    raise ValueError(f'{color} is not a valid 8-bit terminal color')


# --------------------------------------------------------------------------------------


_RGB256_TO_OKLAB = Callable[[int, int, int], tuple[float, float, float]]

class _LUT:

    def __init__(self) -> None:
        self._ansi: dict[Theme, CoordinateVectorSpec] = {}
        self._rgb: None | CoordinateVectorSpec = None
        self._gray: None | CoordinateVectorSpec = None
        self._convert: None | _RGB256_TO_OKLAB = None

    @property
    def convert(self) -> _RGB256_TO_OKLAB:
        if self._convert is None:
            setattr(self, '_convert', get_converter('rgb256', 'oklab'))
        assert self._convert is not None
        return self._convert

    @property
    def ansi(self) -> CoordinateVectorSpec:
        theme = current_theme()
        if theme not in self._ansi:
            self._ansi[theme] = tuple(
                (self.convert if c.tag == 'srgb256' else get_converter(c.tag, 'oklab'))
                (*c.coordinates)  # type: ignore
                for n, c in theme.colors() if n not in ('text', 'background')
            )
        return self._ansi[theme]

    @property
    def rgb(self) -> CoordinateVectorSpec:
        if self._rgb is None:
            self._rgb = tuple(
                self.convert(*rgb6_to_rgb256(r, g, b))
                for r in range(6) for g in range(6) for b in range(6)
            )
        return self._rgb

    @property
    def gray(self) -> CoordinateVectorSpec:
        if self._gray is None:
            self._gray = tuple(
                self.convert(*_eight_bit_gray_to_rgb256(c))
                for c in range(232, 256)
            )
        return self._gray

_look_up_table = _LUT()


def oklab_to_eight_bit(L: float, a: float, b: float) -> tuple[int]:
    """
    :bdg-warning:`Lossy conversion` Convert the given color from Oklab to an
    8-bit terminal color.
    """
    index, _ = closest_oklab((L, a, b), chain(_look_up_table.rgb, _look_up_table.gray))
    return 16 + index,


def oklab_to_rgb6(L: float, a: float, b: float) -> tuple[int, int, int]:
    """
    :bdg-warning:`Lossy conversion` Convert the given color from Oklab to RGB6.
    """
    index, _ = closest_oklab((L, a, b), _look_up_table.rgb)
    return eight_bit_to_rgb6(16 + index)


def oklab_to_ansi(L: float, a: float, b: float) -> tuple[int]:
    """
    :bdg-warning:`Lossy conversion` Convert the given color from Oklab to the
    extended sixteen ANSI colors.

    .. warning::
        The result of this function critically depends on the current color
        theme. It provides an implicit input in addition to the arguments.
    """
    index, _ = closest_oklab((L, a, b), _look_up_table.ansi)
    return index,



from dataclasses import dataclass
import math
import sys
from typing import cast
import unittest

from prettypretty.color.conversion import (
    rgb256_to_srgb,
    srgb_to_rgb256,
    srgb_to_linear_srgb,
    linear_srgb_to_srgb,
    linear_srgb_to_xyz,
    xyz_to_linear_srgb,
    xyz_to_linear_p3,
    linear_p3_to_xyz,
    linear_p3_to_p3,
    p3_to_linear_p3,
    xyz_to_oklab,
    oklab_to_xyz,
    oklab_to_oklch,
    oklch_to_oklab,
)

from prettypretty.color.serde import (
    parse_hex,
    parse_x_rgb,
    parse_x_rgbi,
)

from prettypretty.color.theme import VGA
from prettypretty.api import Color


@dataclass(frozen=True)
class ColorValues:
    spec: str
    parsed: tuple[int, int, int]
    srgb: tuple[float, float, float]
    linear_srgb: tuple[float, float, float]
    p3: tuple[float, float, float]
    linear_p3: tuple[float, float, float]
    xyz: tuple[float, float, float]
    oklab: tuple[float, float, float]
    oklch: tuple[float, float, float]
    ansi: tuple[int]
    black_text: bool
    black_background: bool
    closest_index: int
    x: str
    css: str


class TestColor(unittest.TestCase):

    BLACK = ColorValues(
        spec = '#000',
        parsed = (0, 0, 0),
        srgb = (0.0, 0.0, 0.0),
        linear_srgb = (0.0, 0.0, 0.0),
        p3 = (0.0, 0.0, 0.0),
        linear_p3 = (0.0, 0.0, 0.0),
        xyz = (0.0, 0.0, 0.0),
        oklab = (0.0, 0.0, 0.0),
        oklch = (0.0, 0.0, math.nan),
        ansi = (0,),
        black_text = False,
        black_background = False,
        closest_index = 0,
        x = 'rgb:00/00/00',
        css = 'rgb(0 0 0)',
    )

    YELLOW = ColorValues(
        spec = '#ffca00',
        parsed = (255, 202, 0),
        srgb = (1.0, 0.792156862745098, 0.0),
        linear_srgb = (1.0, 0.5906188409193369, 0.0),
        p3 = (0.967346220711791, 0.8002244967941964, 0.27134084647161244),
        linear_p3 = (0.9273192749713864, 0.6042079205196976, 0.059841923211596565),
        xyz = (0.6235868473237722, 0.635031101987136, 0.08972950140152941),
        oklab = (0.8613332073307732, 0.0017175723640959761, 0.1760013937170005),
        oklch = (0.8613332073307732, 0.17600977428868128, 89.440876452466),
        ansi = (11,),
        black_text = True,
        black_background = True,
        closest_index = 1,
        x = 'rgb:ff/ca/00',
        css = 'rgb(255 202 0)',
    )

    BLUE = ColorValues(
        spec = '#3178ea',
        parsed = (49, 120, 234),
        srgb = (0.19215686274509805, 0.47058823529411764, 0.9176470588235294),
        linear_srgb = (0.030713443732993635, 0.18782077230067787, 0.8227857543962835),
        p3 = (0.2685153556355094, 0.464457615084287, 0.8876966971452301),
        linear_p3 = (0.058605969547446096, 0.18260572039525874, 0.763285235993837),
        xyz = (0.22832473003420622, 0.20025321836938537, 0.80506528557483),
        oklab = (0.5909012953108558, -0.03348086515869708, -0.1836287492414714),
        oklch = (0.5909012953108558, 0.18665606306724153, 259.66681920272583),
        ansi = (12,),
        black_text = False,
        black_background = False,
        closest_index = 2,
        x = 'rgb:31/78/ea',
        css = 'rgb(49 120 234)',
    )

    WHITE = ColorValues(
        spec = '#fff',
        parsed = (255, 255, 255),
        srgb = (1.0, 1.0, 1.0),
        linear_srgb = (1.0, 1.0, 1.0),
        p3 = (0.9999999999999999, 0.9999999999999997, 0.9999999999999999),
        linear_p3 = (1, 0.9999999999999998, 1),
        xyz = (0.9504559270516717, 1.0, 1.0890577507598784),
        oklab = (1.0000000000000002, -4.996003610813204e-16, 1.734723475976807e-17),
        oklch = (1.0000000000000002, 4.999014376318559e-16, math.nan),
        ansi = (15,),
        black_text = True,
        black_background = True,
        closest_index = 3,
        x = 'rgb:ff/ff/ff',
        css = 'rgb(255 255 255)',
    )


    def assertCloseEnough(
        self,
        color1: tuple[float, float, float],
        color2: tuple[float, float, float],
        places: int = 14,
    ) -> None:
        for c1, c2 in zip(color1, color2):
            self.assertAlmostEqual(c1, c2, places=places)


    def test_conversions(self) -> None:
        for color_name in ('BLACK', 'YELLOW', 'BLUE', 'WHITE'):
            values = getattr(self, color_name)
            spec = values.spec

            color_name = color_name.lower()
            with self.subTest('hex-string to RGB256', color=color_name):
                _, rgb256 = parse_hex(spec)
                self.assertTupleEqual(rgb256, values.parsed)

            with self.subTest('RGB256 to sRGB', color=color_name):
                srgb = rgb256_to_srgb(*rgb256)
                self.assertTupleEqual(srgb, values.srgb)

            with self.subTest('sRGB back to RGB256', color=color_name):
                self.assertTupleEqual(srgb_to_rgb256(*srgb), rgb256)

            with self.subTest('sRGB to linear sRGB', color=color_name):
                linear_srgb = srgb_to_linear_srgb(*srgb)
                self.assertTupleEqual(linear_srgb, values.linear_srgb)

            with self.subTest('linear sRGB back to sRGB', color=color_name):
                self.assertCloseEnough(linear_srgb_to_srgb(*linear_srgb), srgb)

            with self.subTest('linear sRGB to XYZ', color=color_name):
                xyz = linear_srgb_to_xyz(*linear_srgb)
                self.assertTupleEqual(xyz, values.xyz)

            with self.subTest('XYZ back to linear sRGB', color=color_name):
                self.assertCloseEnough(xyz_to_linear_srgb(*xyz), linear_srgb)

            with self.subTest('XYZ to linear P3', color=color_name):
                linear_p3 = xyz_to_linear_p3(*xyz)
                self.assertTupleEqual(linear_p3, values.linear_p3)

            with self.subTest('linear P3 back to XYZ', color=color_name):
                self.assertCloseEnough(linear_p3_to_xyz(*linear_p3), xyz)

            with self.subTest('linear P3 to P3', color=color_name):
                p3 = linear_p3_to_p3(*linear_p3)
                self.assertTupleEqual(p3, values.p3)

            with self.subTest('P3 back to linear P3', color=color_name):
                self.assertCloseEnough(p3_to_linear_p3(*p3), linear_p3)

            with self.subTest('XYZ to OkLab', color=color_name):
                oklab = xyz_to_oklab(*xyz)
                self.assertTupleEqual(oklab, values.oklab)

            with self.subTest('OkLab back to XYZ', color=color_name):
                self.assertCloseEnough(oklab_to_xyz(*oklab), xyz)

            with self.subTest('OkLab to OkLCh', color=color_name):
                oklch = oklab_to_oklch(*oklab)
                expected = values.oklch
                self.assertEqual(oklch[0], expected[0])
                self.assertEqual(oklch[1], expected[1])
                self.assertTrue(
                    (math.isnan(oklch[2]) and math.isnan(expected[2]))
                    or oklch[2] == expected[2]
                )

            with self.subTest('OkLCh to OkLab', color=color_name):
                self.assertCloseEnough(oklch_to_oklab(*oklch), oklab)


    def test_lores(self) -> None:
        self.assertFalse('prettypretty.color.lores' in sys.modules)
        cache = sys.modules['prettypretty.color.conversion']._converter_cache
        self.assertEqual(len(cache['ansi']), 0)

        for color_name in ('BLACK', 'YELLOW', 'BLUE', 'WHITE'):
            values = getattr(self, color_name)

            color_name = color_name.lower()
            with self.subTest('lores conversion', color=color_name):
                color = Color(values.spec)
                self.assertEqual(color.tag, 'rgb256')
                self.assertTupleEqual(color.coordinates, values.parsed)

                ansi = color.to('ansi')
                self.assertEqual(ansi.tag, 'ansi')
                self.assertEqual(ansi.coordinates, values.ansi)

                rgb256 = ansi.to('rgb256')
                self.assertTupleEqual(
                    rgb256.coordinates,
                    VGA.ansi(cast(int, ansi.coordinates[0])).coordinates
                )

        self.assertTrue('prettypretty.color.lores' in sys.modules)
        self.assertEqual(len(cache['ansi']), 3)


    def test_color_object(self) -> None:
        for color_name in ('BLACK', 'YELLOW', 'BLUE', 'WHITE'):
            values = getattr(self, color_name)
            spec = values.spec

            color_name = color_name.lower()
            with self.subTest('parse color', color=color_name):
                color = Color(spec)
                self.assertEqual(color.tag, 'rgb256')
                self.assertEqual(color.coordinates, values.parsed)

            with self.subTest('convert to OkLab', color=color_name):
                oklab = color.to('oklab')
                self.assertEqual(oklab.tag, 'oklab')
                self.assertEqual(oklab.coordinates, values.oklab)

            with self.subTest('check distance', color=color_name):
                self.assertEqual(
                    oklab.closest([
                        Color('#444'),
                        Color('#bb0'),
                        Color('#00b'),
                        Color('#bbb'),
                    ]),
                    values.closest_index
                )

            with self.subTest('check contrast', color=color_name):
                self.assertEqual(oklab.use_black_text(), values.black_text)
                self.assertEqual(oklab.use_black_background(), values.black_background)

            with self.subTest('check string', color=color_name):
                self.assertEqual(f'{color:x}', values.x)
                self.assertEqual(f'{color:s}', values.css)


    def test_x_parse_color(self) -> None:
        for text, expected in {
            'rgb:0/8/80': (0, 0x88, 0x80),
            'rgb:800/8000/0': (0.5, 0.5, 0.0),
            'rgbi:0/0.5/1': (0.0, 0.5, 1.0),
            'rgbi:1e0/1e-1/1e-2': (1.0, 0.1, 0.01),
        }.items():
            fn = parse_x_rgb if text.startswith('rgb:') else parse_x_rgbi
            _, actual = fn(text)
            self.assertEqual(actual, expected)



from collections.abc import Iterable
from typing import cast, Literal, overload, Self

from ..color.apca import contrast, use_black_text, use_black_background
from ..color.conversion import get_converter
from ..color.difference import deltaE_oklab, closest_oklab
from ..color.serde import (
    parse_format_spec,
    parse_hex,
    parse_x_rgb,
    parse_x_rgbi,
    stringify,
)
from ..color.space import EPSILON, is_tag, resolve, Space
from ..color.spec import ColorSpec, CoordinateSpec, FloatCoordinateSpec


class Color(ColorSpec):
    """
    A color object.

    This class implements the high-level, object-oriented API for colors.

    Attributes:
        tag: identifies the color format or space
        coordinates: are the numerical components of the color

    Color's constructor supports a number of options for specifying the color
    and its coordinates:

        * From an existing :class:`.ColorSpec` or ``Color`` object
        * From the textual representation of a color, using the ``#``,
          ``rgb:``, or ``rgbi:`` prefix
        * From a tag and tuple with coordinates
        * From a tag and one or three coordinates

    When invoked on an existing color specification or color object, the
    constructor returns a new color object with the same tag and coordinates.
    Hence, it effectively upgrades color specifications to full featured color
    objects.

    Like its parent class :class:`.ColorSpec`, this class is immutable.
    """
    __slots__ = ()

    @property
    def space(self) -> Space:
        """Get the color space for this color."""
        return resolve(self.tag)


    def __getattr__(self, name: str) -> float | Self:
        """
        Provide access to color space coordinates by single-letter name. Also,
        convert to named color format or space.
        """
        if len(name) == 1:
            for coordinate, value in zip(self.space.coordinates, self.coordinates):
                if coordinate.name == name:
                    return value
        elif is_tag(name):
            return self.to(name)

        raise AttributeError(f'color {self} has no attribute named "{name}"')

    # ----------------------------------------------------------------------------------


    @overload
    def __init__(self, tag: str | ColorSpec | Self) -> None:
        ...

    @overload
    def __init__(self, tag: str, coordinates: CoordinateSpec) -> None:
        ...

    @overload
    def __init__(self, tag: str, coordinates: float, c2: float, c3: float) -> None:
        ...

    def __init__(
        self,
        tag: str | ColorSpec | Self,
        coordinates: None | float | CoordinateSpec = None,
        c2: None | float = None,
        c3: None | float = None,
    ) -> None:
        if isinstance(tag, ColorSpec):
            tag, coordinates = tag.tag, tag.coordinates
        elif coordinates is None:
            if tag.startswith('#'):
                tag, coordinates = parse_hex(tag)
            elif tag.startswith('rgb:'):
                tag, coordinates = parse_x_rgb(tag)
            elif tag.startswith('rgbi:'):
                tag, coordinates = parse_x_rgbi(tag)
            else:
                raise ValueError(f'"{tag}" is not a valid color')
        elif c2 is not None:
            assert isinstance(coordinates, float) and c3 is not None
            coordinates = coordinates, c2, c3

        object.__setattr__(self, 'tag', tag)
        object.__setattr__(self, 'coordinates', coordinates)
        self.__post_init__()


    def update(
        self,
        c1: None | float,
        c2: None | float = None,
        c3: None | float = None,
    ) -> Self:
        """
        Update this color.

        Since color objects are immutable, this method returns a new color. That
        color has the same color format or space as this color. Its coordinates
        are the arguments to this method, except that this color's coordinates
        fill null arguments.
        """
        if len(self.coordinates) == 1:
            assert c2 is None and c3 is None
            return type(self)(self.tag, cast(tuple[int], (c1 or self.coordinates[0],)))

        b1, b2, b3 = self.coordinates
        return type(self)(self.tag, (c1 or b1, c2 or b2, c3 or b3))

    # ----------------------------------------------------------------------------------
    # Gamut and Clipping

    def is_in_gamut(self, epsilon: float = EPSILON) -> bool:
        """Determine whether this color is within gamut for its color space."""
        return self.space.is_in_gamut(*self.coordinates, epsilon)


    def clip(self, epsilon: float = 0) -> Self:
        """Clip this color to its color space's gamut."""
        if self.is_in_gamut(epsilon):
            return self
        return type(self)(self.tag, self.space.clip(*self.coordinates))

    # ----------------------------------------------------------------------------------
    # Conversion to Other Formats and Color Spaces

    def to(self, target: str) -> Self:
        """Convert this color to the specified color format or space."""
        if self.tag == target:
            return self
        return type(self)(target, get_converter(self.tag, target)(*self.coordinates))

    # ----------------------------------------------------------------------------------
    # Distance from Other Colors

    def distance(self, other: Self, *, version: Literal[1, 2] = 1) -> float:
        """
        Determine the symmetric distance ΔE between this color and the given
        color.
        """
        return deltaE_oklab(
            *self.to('oklab').coordinates,
            *other.to('oklab').coordinates,
            version=version,
        )

    def closest(self, colors: Iterable[Self]) -> int:
        """
        Find the color with the smallest symmetric distance from this color and
        return its index.
        """
        index, _ = closest_oklab(
            cast(FloatCoordinateSpec, self.to('oklab').coordinates),
            (cast(FloatCoordinateSpec, c.to('oklab').coordinates) for c in colors),
        )
        return index

    # ----------------------------------------------------------------------------------
    # Contrast Against Other Colors

    def contrast_against(self, background: Self) -> float:
        """
        Determine the asymmetric contrast of text with this color against the
        given background.
        """
        return contrast(
            cast(FloatCoordinateSpec, self.to('srgb').coordinates),
            cast(FloatCoordinateSpec, background.to('srgb').coordinates),
        )

    def use_black_text(self) -> bool:
        """
        Determine whether to use black or white text against a background of
        this color for maximum contrast.
        """
        return use_black_text(*self.to('srgb').coordinates)


    def use_black_background(self) -> bool:
        """
        Determine whether to use a black or white background for text of this
        color for maximum contrast.
        """
        return use_black_background(*self.to('srgb').coordinates)

    # ----------------------------------------------------------------------------------
    # Serialization to Text

    def __format__(self, format_spec: str) -> str:
        fmt, precision = parse_format_spec(format_spec)
        return stringify(self.tag, self.coordinates, fmt, precision)


    def __str__(self) -> str:
        return stringify(self.tag, self.coordinates)



"""
Metadata about color spaces.

Note that the ``conversion`` and ``serde`` modules have local copies of some of
the same data. If you update this module, be sure to update the other two
modules, too.
"""
import dataclasses
import math
from typing import cast, Literal, Self

from .spec import CoordinateSpec


EPSILON = 0.000075


@dataclasses.dataclass(frozen=True, slots=True)
class Coordinate:
    """
    A color space coordinate.

    Attributes:
        name: the single-letter name of the coordinate or an empty string
            for nameless coordinates
        min: the optional minimum value for the coordinate
        max: the optional maximum value for the coordinate
        type: the optional type for common coordinate semantics

    The following three types are recognized:

      * An **angle** is a floating point number representing a rotation between
        0 and 360 degrees, inclusive; used for the *hue* in polar color spaces.
      * A **normal** is a floating point number between 0 and 1, inclusive.
      * An **int** may have arbitrary range but all its values must be integral.
    """
    name: str
    min: None | float = None
    max: None | float = None
    type: None | Literal['angle', 'int', 'normal'] = None

    def __post_init__(self) -> None:
        if len(self.name) > 1:
            raise ValueError('coordinate must have name with at most 1 character')
        elif self.type == 'angle':
            if self.min != 0 or self.max != 360:
                raise ValueError('angle coordinate must have range from 0 to 360')
        elif self.type == 'normal':
            if self.min != 0 or self.max != 1:
                raise ValueError('normal coordinate must have range from 0 to 1')

    def is_in_range(self, value: float, *, epsilon: float = EPSILON) -> bool:
        """
        Determine whether the value is within the range set by this coordinate's
        ``min`` and ``max`` attributes with an ``epsilon`` tolerance. If either
        limit is ``None``, that limit is not tested.
        """
        if math.isnan(value):
            return self.type == 'angle'
        if self.type == 'int' and not isinstance(value, int):
            return False
        return (
            (self.min is None or self.min - epsilon <= value)
            and (self.max is None or value <= self.max + epsilon)
        )

    def clip(self, value: float, *, epsilon: float = 0) -> float:
        """
        Clip the value to the range set by this coordinate's ``min`` and ``max``
        attributes with an ``epsilon`` tolerance. If either limits is ``None``,
        that limit has no impact on the value.
        """
        if self.min is not None and value < self.min - epsilon:
            return self.min
        if self.max is not None and value > self.max + epsilon:
            return self.max
        return value


@dataclasses.dataclass(frozen=True, slots=True)
class Space:
    """
    A color format or space.

    Attributes:
        tag: is a lower-case Python identifier
        name: is a human-readable, descriptive label
        base: is the optional base color space
        coordinates: are the coordinates
        css_format: is the optional CSS format
        lores: is the flag for a low-resolution format
    """
    tag: str
    label: str
    base: None | Self
    coordinates: tuple[Coordinate, ...]
    css_format: None | str = None
    lores: bool = False

    def is_in_gamut(self, *coordinates: float, epsilon: float = EPSILON) -> bool:
        """
        Determine whether the coordinates are in gamut for this color space with
        an ``epsilon`` tolerance.
        """
        assert len(self.coordinates) == len(coordinates)
        for coordinate, value in zip(self.coordinates, coordinates):
            if not coordinate.is_in_range(value, epsilon=epsilon):
                return False
        return True

    def clip(self, *coordinates: float, epsilon: float = 0) -> CoordinateSpec:
        """
        Clip the coordinates to this color space's gamut with an ``epsilon``
        tolerance.
        """
        return cast(
            CoordinateSpec,
            tuple(
                c.clip(v, epsilon=epsilon)
                for c, v in zip(self.coordinates, coordinates)
            ),
        )


_RGB_COORDINATES = (
    Coordinate('r', 0, 1, 'normal'),
    Coordinate('g', 0, 1, 'normal'),
    Coordinate('b', 0, 1, 'normal'),
)

XYZ = XYZ_D65 = Space(
    tag='xyz',
    label='XYZ D65',
    base=None,
    coordinates=(
        Coordinate('X'),
        Coordinate('Y'),
        Coordinate('Z'),
    ),
    css_format='color(xyz {})',
)

LINEAR_SRGB = Space(
    tag='linear_srgb',
    label='Linear sRGB',
    base=XYZ,
    coordinates=_RGB_COORDINATES,
    css_format='color(srgb-linear {})',
)

SRGB = Space(
    tag='srgb',
    label='sRGB',
    base=LINEAR_SRGB,
    coordinates=_RGB_COORDINATES,
    css_format='color(srgb {})',
)

RGB256 = Space(
    tag='rgb256',
    label='24-bit RGB',
    base=SRGB,
    coordinates=(
        Coordinate('r', 0, 255, 'int'),
        Coordinate('g', 0, 255, 'int'),
        Coordinate('b', 0, 255, 'int'),
    ),
    css_format='rgb({})',
)

LINEAR_P3 = Space(
    tag='linear_p3',
    label='Linear P3',
    base=XYZ,
    coordinates=_RGB_COORDINATES,
    css_format=None,
)

P3 = Space(
    tag='p3',
    label='P3',
    base=LINEAR_P3,
    coordinates=_RGB_COORDINATES,
    css_format='color(display-p3 {})',
)

OKLAB = Space(
    tag='oklab',
    label='Oklab',
    base=XYZ,
    coordinates=(
        Coordinate('L', 0, 1, 'normal'),
        Coordinate('a', -0.4, 0.4),
        Coordinate('b', -0.4, 0.4),
    ),
    css_format='oklab({})',
)

OKLCH = Space(
    tag='oklch',
    label='Oklch',
    base=OKLAB,
    coordinates=(
        Coordinate('L', 0, 1, 'normal'),
        Coordinate('C', 0, 0.4),
        Coordinate('h', 0, 360, 'angle'),
    ),
    css_format='oklch({})',
)

EIGHT_BIT = Space(
    tag='eight_bit',
    label='8-bit terminal color',
    base=RGB256,
    coordinates=(
        Coordinate('', 0, 255, 'int'),
    ),
    css_format=None,
    lores=True,
)

RGB6 = Space(
    tag='rgb6',
    label='8-bit terminal color RGB',
    base=EIGHT_BIT,
    coordinates=(
        Coordinate('r', 0, 5, 'int'),
        Coordinate('g', 0, 5, 'int'),
        Coordinate('b', 0, 5, 'int'),
    ),
    css_format=None,
    lores=True,
)

ANSI = Space(
    tag='ansi',
    label='ANSI terminal color',
    base=EIGHT_BIT,
    coordinates=(
        Coordinate('', 0, 15, 'int'),
    ),
    css_format=None,
    lores=True,
)

UNIVERSE = (
    # Color spaces
    XYZ,
    LINEAR_SRGB,
    SRGB,
    LINEAR_P3,
    P3,
    OKLAB,
    OKLCH,

    # In-memory format
    RGB256,

    # Terminal-specific in-memory format
    RGB6,
    EIGHT_BIT,
    ANSI,
)

_TAG_TO_SPACE = { space.tag: space for space in UNIVERSE }

def resolve(tag: str) -> Space:
    """Resolve the given tag into the corresponding color space metadata."""
    return _TAG_TO_SPACE[tag]

def is_tag(tag: str) -> bool:
    """
    Determine whether the given string is a valid tag for a color format or
    space.
    """
    return tag in _TAG_TO_SPACE



"""
Support for representing the sixteen extended ANSI colors and for mapping them
to meaningful color values.
"""

from collections.abc import Iterator
from contextlib import contextmanager
import dataclasses
from typing import cast, ContextManager, overload

from .spec import ColorSpec, IntCoordinateSpec


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Theme:
    """
    A terminal color theme.

    This class collects color values for the default text (foreground) and
    background colors as well as the 16 extended ANSI colors. While many
    terminal themes affect additional user interface elements, such as cursor
    color, ANSI escape codes cannot modify the appearance of these additional UI
    elements and hence they can be safely ignored.

    Instances of this class are immutable.
    """
    text: ColorSpec
    background: ColorSpec
    black: ColorSpec
    red: ColorSpec
    green: ColorSpec
    yellow: ColorSpec
    blue: ColorSpec
    magenta: ColorSpec
    cyan: ColorSpec
    white: ColorSpec
    bright_black: ColorSpec
    bright_red: ColorSpec
    bright_green: ColorSpec
    bright_yellow: ColorSpec
    bright_blue: ColorSpec
    bright_magenta: ColorSpec
    bright_cyan: ColorSpec
    bright_white: ColorSpec

    def colors(self) -> Iterator[tuple[str, ColorSpec]]:
        """Get an iterator over the name, color pairs of this theme."""
        for field in dataclasses.fields(self):
            yield field.name, getattr(self, field.name)

    def ansi(self, color: int) -> ColorSpec:
        """Look up the RGB256 coordinates for the extended ANSI color."""
        assert 0 <= color <= 15
        return getattr(self, dataclasses.fields(self)[2 + color].name)


def hex(color: str) -> ColorSpec:
    """
    Convert a string with exactly six hexadecimal digits into an RGB256 color.
    """
    return ColorSpec(
        'rgb256',
        cast(
            IntCoordinateSpec,
            tuple(int(color[n : n+2], base=16) for n in range(0, 6, 2)),
        ),
    )

MACOS_TERMINAL = Theme(
    text=hex("000000"),
    background=hex("ffffff"),
    black=hex("000000"),
    red=hex("990000"),
    green=hex("00a600"),
    yellow=hex("999900"),
    blue=hex("0000b2"),
    magenta=hex("b200b2"),
    cyan=hex("00a6b2"),
    white=hex("bfbfbf"),
    bright_black=hex("666666"),
    bright_red=hex("e50000"),
    bright_green=hex("00d900"),
    bright_yellow=hex("e5e500"),
    bright_blue=hex("0000ff"),
    bright_magenta=hex("e500e5"),
    bright_cyan=hex("00e5e5"),
    bright_white=hex("e5e5e5"),
)


VGA = Theme(
    text=hex("000000"),
    background=hex("ffffff"),
    black=hex("000000"),
    red=hex("aa0000"),
    green=hex("00aa00"),
    yellow=hex("aa5500"),
    blue=hex("0000aa"),
    magenta=hex("aa00aa"),
    cyan=hex("00aaaa"),
    white=hex("aaaaaa"),
    bright_black=hex("555555"),
    bright_red=hex("ff5555"),
    bright_green=hex("55ff55"),
    bright_yellow=hex("ffff55"),
    bright_blue=hex("5555ff"),
    bright_magenta=hex("ff55ff"),
    bright_cyan=hex("55ffff"),
    bright_white=hex("ffffff"),
)

XTERM = Theme(
    text=hex("000000"),
    background=hex("ffffff"),
    black=hex("000000"),
    red=hex("cd0000"),
    green=hex("00cd00"),
    yellow=hex("cdcd00"),
    blue=hex("0000ee"),
    magenta=hex("cd00cd"),
    cyan=hex("00cdcd"),
    white=hex("e5e5e5"),
    bright_black=hex("7f7f7f"),
    bright_red=hex("ff0000"),
    bright_green=hex("00ff00"),
    bright_yellow=hex("ffff00"),
    bright_blue=hex("5c5cff"),
    bright_magenta=hex("ff00ff"),
    bright_cyan=hex("00ffff"),
    bright_white=hex("ffffff"),
)


_current_theme = VGA

@overload
def current_theme() -> Theme:
    ...
@overload
def current_theme(theme: Theme) -> ContextManager[Theme]:
    ...
def current_theme(theme: None | Theme = None) -> Theme | ContextManager[Theme]:
    """
    Manage the current theme.

    This function does the work of two:

     1. When invoked without arguments, this function simply returns the current
        theme.
     2. When invoked with a theme as argument, this function returns a context
        manager that switches to the provided theme on entry and restores the
        current theme again on exit.

    The default theme uses the same colors as good ol' VGA text mode.
    """
    if theme is None:
        return _current_theme

    @contextmanager
    def another_theme() -> Iterator[Theme]:
        global _current_theme
        saved_theme = _current_theme
        _current_theme = theme
        try:
            yield theme
        finally:
            _current_theme = saved_theme

    return another_theme()
